/*
 *  tcp.c  --  telnet protocol communication module for powwow
 *
 *  Copyright (C) 1998 by Massimiliano Ghilardi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/telnet.h>
#ifndef TELOPT_NAWS
#  define TELOPT_NAWS 31
#endif
#include <arpa/inet.h>
#ifndef NEXT
#  include <unistd.h>
#endif

#ifdef TERM
#  include "client.h"
#endif

#include "defines.h"
#include "main.h"
#include "utils.h"
#include "tcp.h"
#include "tty.h"
#include "edit.h"
#include "beam.h"
#include "log.h"

#ifdef TELOPTS
# define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

int tcp_fd = -1;                   /* current socket file descriptor
				      -1 means no socket */
int tcp_main_fd = -1;		   /* socket file descriptor of main connect.
				      -1 means no socket */
int tcp_count = 0;		   /* number of open connections */
int tcp_attachcount = 0;	   /* number of spawned commands */

connsess conn_list[MAX_FDSCAN];    /* connection list */

fd_set fdset;			/* set of descriptors to select() on */

/*
 * process suboptions.
 * so far, only terminal type is processed but future extensions are
 * window size, X display location, etc.
 */
static void dosubopt __P1 (byte *,str)
{
    char buf[256], *term;
    int len, err;

    if (str[0] == TELOPT_TTYPE) {
	if (str[1] == 1) {
	    /* 1 == SEND */
#ifdef TELOPTS
	    tty_printf("[got SB TERMINAL TYPE SEND]\n");
#endif
	    if (!(term = getenv("TERM"))) term = "unknown";
	    sprintf(buf, "%c%c%c%c%.*s%c%c", IAC, SB, TELOPT_TTYPE, 0,
		    256-7, term, IAC, SE);	/* 0 == IS */
	    
	    len = strlen(term) + 6;
	    while ((err = write(tcp_fd, buf, len)) < 0 && errno == EINTR)
	        ;
	    if (err != len) {
	        errmsg("write subopt to socket");
		return;
	    }
#ifdef TELOPTS
	    tty_printf("[sent SB TERMINAL TYPE IS %s]\n", term);
#endif
	}
    }
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void sendopt __P2 (byte,what, byte,opt)
{
    static byte buf[3] = { IAC, 0, 0 };
    int i;
    buf[1] = what; buf[2] = opt;
    
    while ((i = write(tcp_fd, buf, 3)) < 0 && errno == EINTR)
	;
    if (i != 3) {
	errmsg("write option to socket");
	return;
    }

#ifdef TELOPTS
    tty_printf("[sent %s %s]\n", (what == WILL) ? "WILL" :
	       (what == WONT) ? "WONT" :
	       (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	       TELOPTSTR(opt));
#endif
}

/*
 * connect to remote host
 * Warning: some voodoo code here
 */
int tcp_connect __P2 (char *,addr, int,port)
{
    struct sockaddr_in address;
    struct hostent *host_info;
    int err, newtcp_fd;
    
    status(1);
    
    memzero((char *)&address, sizeof(address));
    /*
     * inet_addr has a strange design: It is documented to return -1 for
     * malformed requests, but it is declared to return unsigned long!
     * Anyway, this works.
     */    
    
#ifndef TERM  
    address.sin_addr.s_addr = inet_addr(addr);
    if (address.sin_addr.s_addr != (unsigned int)-1)
	address.sin_family = AF_INET;
    else
    {
	if (echo_int)
	    tty_printf("#looking up %s... ", addr);
	tty_flush();
	host_info = gethostbyname(addr);
	if (host_info == 0) {
	    if (!echo_int) {
		tty_printf("#looking up %s... ", addr);
	    }
	    tty_printf("unknown host!\n");
	    return -1;
	}
	memmove((char *)&address.sin_addr, host_info->h_addr,
	      host_info->h_length);
	address.sin_family = host_info->h_addrtype;
	if (echo_int)
	    tty_puts("found.\n");
    }
    address.sin_port = htons(port);
    
    newtcp_fd = socket(address.sin_family, SOCK_STREAM, 0);
    if (newtcp_fd == -1) {
	errmsg("create socket");
	return -1;
    } else if (newtcp_fd >= MAX_FDSCAN) {
	tty_printf("#connect: #error: too many open connections\n");
	close(newtcp_fd);
	return -1;
    }
    
    tty_printf("#trying %s... ", addr);
    tty_flush();
    
    err = connect(newtcp_fd, (struct sockaddr *)&address, sizeof(address));
    
    if (err == -1) { /* CTRL-C pressed, or other errors */
	errmsg("connect to host");
	return -1;
    }
    
#else /* term */
    
    if ((newtcp_fd = connect_server(0)) < 0) {
	tty_puts("\n#powwow: unable to connect to term server\n");
	return -1;
    } else {
	if (newtcp_fd >= MAX_FDSCAN) {
	    tty_printf("#connect: #error: too many open connections\n");
	    close(newtcp_fd);
	    return -1;
	}
	send_command(newtcp_fd, C_PORT, 0, "%s:%d", addr, port);
	tty_puts("Connected to term server...\n");
#ifdef TERM_COMPRESS
	send_command(newtcp_fd, C_COMPRESS, 1, "y");
#endif     
	send_command(newtcp_fd, C_DUMB, 1, 0);
    }

#endif /* term */
    
    tty_puts("connected!\n");
    
    
    {
	/*
	 * Now set some options on newtcp_fd :
	 * First, no-nagle
	 */
	int opt = 1;
#	ifndef SOL_TCP
#	 define SOL_TCP IPPROTO_TCP
#	endif
	if (setsockopt(newtcp_fd, SOL_TCP, TCP_NODELAY, &opt, sizeof(opt)))
	    errmsg("setsockopt(TCP_NODELAY) failed");
	
	/*
	 * Then, close-on-exec:
	 * we don't want children to inherit the socket!
	 */
	
	fcntl(newtcp_fd, F_SETFD, FD_CLOEXEC);
    }
    
    return newtcp_fd;
}

/*
 * read a maximum of size chars from remote host
 * using the telnet protocol. return chars read.
 */
int tcp_read __P3 (int,fd, char *,buffer, int,maxsize)
{
    typedef enum {
	NORMAL = 0, GOTIAC, GOTWILL, GOTWONT, GOTDO, GOTDONT, GOTSB, GOTSBIAC
    } telnetstate;
    static telnetstate sock_state[MAX_FDSCAN];
    telnetstate state = sock_state[fd];
    int i;
    static byte subopt[MAX_SUBOPT];
    static int subchars;
    byte *p, *s;
    
    while ((i = read(fd, buffer, maxsize)) < 0 && errno == EINTR)
	;
    
    if (i <= 0) {
	if (i == 0) {
	    sock_state[fd] = NORMAL;
	    tcp_close(NULL);
	    return 0;
	}
	errmsg("read from socket");
	return 0;
    }
    /* scan through the buffer and interpret telnet protocol escapes */
    
    for (s = p = (byte *)buffer; i--; s++) {
	switch (state) {
	  case NORMAL:
	  default:
	    if (*s != IAC)
		/*
		 * Some servers like to send NULs. Replace them with spaces,
		 * and hope for the best (I really should cut them out entirely)
		 */
		*p++ = *s ? *s : ' ';
	    else
		state = GOTIAC;
	    break;
	  case GOTIAC:
	    switch(*s) {
	      case WILL:
		state = GOTWILL; break;
	      case WONT:
		state = GOTWONT; break;
	      case DO:
		state = GOTDO; break;
	      case DONT:
		state = GOTDONT; break;
	      case SB:		/* BUG (multiple connections):	*/
		state = GOTSB;	/* there is only one subopt buffer */
		subchars = 0;
		break;
	      case IAC:
		*p++ = IAC;
		break;
	      case GA:
		/* I should handle GA as end-of-prompt marker one day */
	      default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		tty_printf("[skipped IAC <%d>]\n", *s);
#endif
		state = NORMAL; break;
	    }
	    break;
	    
	  case GOTWILL:
#ifdef TELOPTS
	    tty_printf("[got WILL %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	      case TELOPT_ECHO:
		/* host echoes, turn off echo here
		 * but only for main connection, since we do not want
		 * subsidiary connection password entries to block anything
		 * in the main connection
		 */
		if (fd == tcp_main_fd)
		    linemode |= LM_NOECHO;
		sendopt(DO, *s);
		break;
	      case TELOPT_SGA:
		/* this can't hurt */
		linemode |= LM_CHAR;
		tty_special_keys();
		sendopt(DO, *s);
		break;
	      default:
		/* don't accept other options */
		sendopt(DONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;
	    
	  case GOTWONT:
#ifdef TELOPTS
	    tty_printf("[got WONT %s]\n", TELOPTSTR(*s));
#endif
	    if (*s == TELOPT_ECHO) {
		/* host no longer echoes, we do it instead */
		linemode &= ~LM_NOECHO;
	    }
	    /* accept any WONT */
	    sendopt(DONT, *s);
	    state = NORMAL;
	    break;
	    
	  case GOTDO:
#ifdef TELOPTS
	    tty_printf("[got DO %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	      case TELOPT_SGA:
		linemode |= LM_CHAR;
		tty_special_keys();
		/* FALLTHROUGH */
	      case TELOPT_TTYPE:
		sendopt(WILL, *s);
		break;
	      case TELOPT_NAWS:
		sendopt(WILL, *s);
		tcp_write_tty_size();
		break;
	      default:
		/* accept nothing else */
		sendopt(WONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;
	    
	  case GOTDONT:
#ifdef TELOPTS
	    tty_printf("[got DONT %s]\n", TELOPTSTR(*s));
#endif
	    if (*s == TELOPT_SGA) {
		linemode &= ~LM_CHAR;
		tty_special_keys();
	    }
	    sendopt(WONT, *s);
	    state = NORMAL;
	    break;
	    
	  case GOTSB:
	    if (*s == IAC) {
		state = GOTSBIAC;
	    } else {
		if (subchars < MAX_SUBOPT)
		    subopt[subchars++] = *s;
	    }
	    break;
	    
	  case GOTSBIAC:
	    if (*s == IAC) {
		if (subchars < MAX_SUBOPT)
		    subopt[subchars++] = IAC;
		state = GOTSB;
	    } else if (*s == SE) {
		subopt[subchars] = '\0';
		dosubopt(subopt);
		state = NORMAL;
	    } else {
		/* problem! I haven't the foggiest idea of what to do here.
		 * I'll just ignore it and hope it goes away. */
		PRINTF("#telnet: got IAC <%d> instead of IAC SE!\n", (int)*s);
		state = NORMAL;
	    }
	    break;
	}
    }
    sock_state[fd] = state;
    
    log_write(buffer, (char *)p - buffer, 0);
    
    return (char *)p - buffer;
}

void tcp_raw_write __P3 (int,fd, char *,data, int,len)
{
    int i;
    tcp_flush();
    while (len > 0) {
	while ((i = write(fd, data, len)) < 0 && errno == EINTR)
	    ;
	if (i < 0) {
	    errmsg("write to socket");
	    break;
	}
	data += i;
	len -= i;
    }
}

/*
 * Send current terminal size (RFC 1073)
 */
void tcp_write_tty_size __P0 (void)
{
    static byte buf[] = { IAC, SB, TELOPT_NAWS, 0, 0, 0, 0, IAC, SE };
    
    buf[3] = cols >> 8;
    buf[4] = cols & 0xff;
    buf[5] = lines >> 8;
    buf[6] = lines & 0xff;
    
    tcp_raw_write(tcp_main_fd, (char *)buf, 9);
#ifdef TELOPTS
    tty_printf("[sent term size %d %d]\n", cols, lines);
#endif
}

/*
 * send a string to the main connection on the remote host
 */
void tcp_main_write __P1 (char *,data)
{
    tcp_write(tcp_main_fd, data);
}


static char output_buffer[BUFSIZE];
static int output_len = 0;	/* number of characters in output_buffer */
static int output_socket = -1;	/* to which socket buffer should be sent*/

/*
 * put data in the output buffer for transmission to the remote host
 */
void tcp_write __P2 (int,fd, char *,data)
{
    int len, space;
    len = strlen(data);
    
    if (tcp_main_fd != -1 && tcp_main_fd == fd) {
	if (linemode & LM_NOECHO)
	    log_write("", 0, 1); /* log a newline only */
	else
	    log_write(data, strlen(data), 1);
	reprint_writeline(data);
    }

    status(-1); /* must be AFTER reprint_writeline() */

    if (fd != output_socket) { /* is there data to another socket? */
	tcp_flush(); /* then flush it */
	output_socket = fd;
    }
    space = BUFSIZE - output_len;
    if (space > len) {  		/* enough room? */
	memmove(output_buffer + output_len, data, len);
	output_buffer[ output_len += len ] = '\n';
	output_len++;
    } else {
	memmove(output_buffer + output_len, data, space);
	output_len = BUFSIZE;
	tcp_flush();
	tcp_write(fd, data + space);
    }
}

/*
 * send all buffered data to the remote host
 */
void tcp_flush __P0 (void)
{
    int n;
    char *p = output_buffer;
    
    if (output_len && output_socket == -1) {
	clear_input_line(1);
	PRINTF("#no open connections. Use '#connect main <address> <port>' to open a connection.\n");
	output_len = 0;
	return;
    }
    
    if (!output_len)
	return;
    
    while (output_len) {	
	while ((n = write(output_socket, p, output_len)) < 0 && errno == EINTR)
	    ;
	if (n < 0) {
	    errmsg("write to socket");
	    return;
	}
	sent += n;
	p += n;
	output_len -= n;
    }
    
    status(-1); /* sent stuff, so we expect a prompt */
}

/*
 * Below are multiple-connection support functions:
 */

/*
 * return connection's index given id,
 * or -1 if null or invalid id is given
 */
int tcp_find __P1 (char *,id)
{
    int i, n;
    
    for (i=0, n=tcp_count+tcp_attachcount; n; i++)
	if (CONN_LIST(i).id) {
	    n--;
	    if (!strcmp(CONN_LIST(i).id, id))
		return i;
	}
    return -1;
}

/*
 * show list of open connections
 */
void tcp_show __P0 (void)
{
    int i, n=tcp_count+tcp_attachcount;
    
    PRINTF("#%s connection%s opened%c\n", n ? "The following" : "No", 
	       n==1 ? " is" : "s are", n ? ':' : '.');
    
    
    for (i=0, n=tcp_count; n; i++)
	if (CONN_LIST(i).id && CONN_LIST(i).mode == 0) {
	    n--;
	    tty_printf("MUD %sactive %s ##%s\t (%s %d)\n",
		       CONN_LIST(i).active ? "   " : "non",
		       i == tcp_main_fd ? "(default)" : "         ",
		       CONN_LIST(i).id, 
		       CONN_LIST(i).host, CONN_LIST(i).port);
	}
    for (i=0, n=tcp_attachcount; n; i++)
	if (CONN_LIST(i).id && CONN_LIST(i).mode == 1) {
	    n--;
	    tty_printf("CMD %sactive %s ##%s\t (%s)\n",
		       CONN_LIST(i).active ? "   " : "non",
		       i == tcp_main_fd ? "(default)" : "         ",
		       CONN_LIST(i).id, CONN_LIST(i).host);
	}
}

/*
 * permanently change main connection
 */
void tcp_set_main __P1 (int,fd)
{
    /* GH: reset linemode and prompt */
    tcp_main_fd = fd;
    if (linemode & LM_CHAR) 
	linemode = 0, tty_special_keys();
    else
	linemode = 0;
    status(-1);
    reprint_clear();
}

/*
 * open another connection
 */
void tcp_open __P4 (char *,id, char *,initstring, char *,host, int,port)
{
    int newtcp_fd;
    
    if (tcp_find(id)>=0) {
	PRINTF("#connection `%s' already open.\n", id);
	return;
    }
    
    /* dial the number by moving the right index in small circles */
    if ((newtcp_fd = tcp_connect(host, port)) < 0)
	return;
    if (echo_int && tcp_count) {
	PRINTF("#default connection is now `%s'\n", id);
    }
    
    if (!(CONN_LIST(newtcp_fd).id = my_strdup(id))) {
	errmsg("malloc");
	close(newtcp_fd);
	return;
    }
    if (!(CONN_LIST(newtcp_fd).host = my_strdup(host))) {
	errmsg("malloc");
	free(CONN_LIST(newtcp_fd).id);
	close(newtcp_fd);
	return;
    }
    CONN_LIST(newtcp_fd).active = 1;
    CONN_LIST(newtcp_fd).mode = 0;
    CONN_LIST(newtcp_fd).port = port;
	
    FD_SET(newtcp_fd, &fdset); 		/* add socket to select() set */
    tcp_count++;

    tcp_set_main(tcp_fd = newtcp_fd);    
    tcp_write_tty_size();
    
    if (initstring) {
	parse_instruction(initstring, 0, 0, 1);
	history_done = 0;
    }
}

/*
 * close a connection
 */
void tcp_close __P1 (char *,id)
{
    int i, sfd;
    
    status(1);
    tty_puts(edattrend);
    /*
     * because we may be called from get_remote_input()
     * if tcp_read gets an EOF, before edattrend is
     * printed by get_remote_input() itself.
     */
    
    if (id) {  /* #disconnect cmd */
	if ((sfd = tcp_find(id)) < 0) {
	    tty_printf("#no such connection: `%s'\n", id);
	    return;
	}
    } else
	sfd = tcp_fd;  /* connection closed by remote host */

    close(sfd);

    abort_edit_fd(sfd);
    
    tty_printf("#connection on `%s' closed.\n", CONN_LIST(sfd).id);
    
    if (sfd == tcp_main_fd) { /* main connection closed */
	if (tcp_count == 1) { /* was last connection */
	    if (opt_exit)
		exit_powwow();
	    tty_puts("#no connections left. Type #quit to quit.\n");
	    tcp_fd = tcp_main_fd = -1;
	} else {
	    /* must find another connection and promote it to main */
	    for (i = 0; i < MAX_FDSCAN; i++) {
		if (i == sfd)
		    continue;
		if (CONN_LIST(i).id && CONN_LIST(i).mode == 0) {
		    tty_printf("#default connection is now `%s'\n", CONN_LIST(i).id);
		    tcp_main_fd = i;
		    break;
		}
	    }
	    if (sfd == tcp_main_fd) {
	        tty_printf(
		     "#PANIC! internal error in tcp_close()\nQuitting.\n");
		syserr(NULL);
	    }
	}
	tcp_set_main(tcp_main_fd);
    }
    
    if (tcp_fd == sfd)
	tcp_fd = -1; /* no further I/O allowed on sfd, as we just closed it */
    
    FD_CLR(sfd, &fdset);
    if (CONN_LIST(sfd).mode == 0)
	tcp_count--;
    else
	tcp_attachcount--;
    CONN_LIST(sfd).active = 0;
    CONN_LIST(sfd).mode = 0;
    CONN_LIST(sfd).port = 0;
    free(CONN_LIST(sfd).host); CONN_LIST(sfd).host = 0;
    free(CONN_LIST(sfd).id);   CONN_LIST(sfd).id = 0;
}

/*
 * toggle output display from another connection
 */
void tcp_togglesnoop __P1 (char *,id)
{
    int sfd;
    
    sfd = tcp_find(id);
    if (sfd>=0) {
        CONN_LIST(sfd).active ^= 1;
	if (echo_int) {
	    PRINTF("#connection %s is now %sactive.\n",
		      CONN_LIST(sfd).id, CONN_LIST(sfd).active ? "" : "non");
	}
    } else {
        PRINTF("#no such connection: %s\n", id);
    }
}

void tcp_spawn __P2 (char *,id, char *,cmd)
{
    char *buf;
    int childpid, sockets[2];

    if (tcp_find(id)>=0) {
	PRINTF("#connection `%s' already open.\n", id);
	return;
    }
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets) < 0) {
	errmsg("create socketpair");
	return;
    }
    unescape(cmd);
    buf = (char *)malloc(6 + strlen(cmd));
    if (!buf) {
	close(sockets[0]);
	close(sockets[1]);
	errmsg("malloc");
	return;
    }
    sprintf(buf, "exec %s", cmd);
    
    switch (childpid = fork()) {
      case 0:
	/* child */
	close(0); close(1); close(2);
	dup2(sockets[1], 0);
	dup2(sockets[1], 1);
	dup2(sockets[1], 2);
	close(sockets[0]);
	close(sockets[1]);
	execl("/bin/sh", "sh", "-c", cmd, NULL);
	syserr("execl");
	break;
      case -1:
	close(sockets[0]);
	close(sockets[1]);
	free(buf);
	errmsg("fork");
	return;
    }
    close(sockets[1]);
    
    /* Again, we don't want children to inherit sockets */
    fcntl(sockets[0], F_SETFD, FD_CLOEXEC);
    
    if (!(CONN_LIST(sockets[0]).id = my_strdup(id))) {
	errmsg("malloc");
	close(sockets[0]);
	return;
    }
    if (!(CONN_LIST(sockets[0]).host = my_strdup(cmd))) {
	errmsg("malloc");
	free(CONN_LIST(sockets[0]).id);
	close(sockets[0]);
	return;
    }
    CONN_LIST(sockets[0]).active = 1;
    CONN_LIST(sockets[0]).mode = 1;
    CONN_LIST(sockets[0]).port = 0;
    
    FD_SET(sockets[0], &fdset);	       /* add socket to select() set */
    tcp_attachcount++;
    
    if (echo_int) {
	PRINTF("#successfully spawned `%s' with pid %d\n", id, childpid);
    }
    
    /*
     * when the child exits we also get an EOF on the socket,
     * so no special care is needed by the SIGCHLD handler.
     */    
}

